<?php

use WHMCS\Database\Capsule;
use SiteHost\WHMCS\ApiClient;
use WHMCS\View\Menu\Item as MenuItem;

require_once __DIR__ . '/lib/autoloader.php';

function check_invoice_header($invoice_id): bool
{
  $invoice_header = Capsule::table('tblinvoices')->where('id', $invoice_id)->first();
  if (is_null($invoice_header))
  {
	// No invoice found - return
	return false;
  }

  return true;
}

function check_invoice_items($invoice_id)
{
  $invoice_item = Capsule::table('tblinvoiceitems')
						 ->where('invoiceid', $invoice_id)
						 ->where('type', 'DomainTransfer')
						 ->get();

  if (count($invoice_item) <= 0)
  {
	// No invoice found - return
	return false;
  }

  return $invoice_item;
}

function check_nz_domain_rules($params)
{
  try
  {
	$invoice_id = $params['invoiceid'];
	// Check if invoice exists
	if (!check_invoice_header($invoice_id))
	{
	  return;
	}
	// Check if there is an invoice item for this invoice
	if (!$invoice_item = check_invoice_items($invoice_id))
	{
	  return;
	}

	$api_client = new ApiClient();
	foreach ($invoice_item as $item)
	{
	  // Get the domain
	  $domain_row = Capsule::table('tbldomains')
						   ->where('id', $item->relid)
						   ->first();
	  if (is_null($domain_row))
	  {
		continue;
	  }
	  // Check the rules for .NZ domains
	  $should_renew = $api_client->ShouldBeRenewed($domain_row->domain);

	  if ($should_renew)
	  {
		continue;
	  }

	  $old_description = explode(' - ', $item->description);

	  // Update description of the invoice depending on if we have two's hyphens in the text.
	  // Example: "Domain Transfeer - example.co.nz - 1 Year/s"
	  if (is_array($old_description) && count($old_description) >= 2)
	  {
		$new_description = $old_description[0] . ' - ' . $old_description[1];
	  }
	  else
	  {
		// If the criteria for the new description are not meet, we set to the old name in order to not change
		// the description
		$new_description = $item->description;
	  }

	  Capsule::table('tblinvoiceitems')
			 ->where('id', $item->id)
			 ->update([
						  'amount'      => 0.00,
						  'description' => $new_description,
					  ]);

	  Capsule::table('tbldomains')
			 ->where('id', $item->relid)
			 ->update(
				 [
					 'firstpaymentamount' => 0.00,
				 ]);


	  logActivity("[DOMAIN TRANSFER INTERCEPT] Prices for renewal updated - Invoice_id: {$params['invoiceid']} | Domain: {$domain_row->domain}");
	}
  }
  catch (\Exception $e)
  {
	logActivity("[DOMAIN TRANSFER INTERCEPT] - ERROR ATTEMPTING TO UPDATE A RECORD {$e->getMessage()}");
  }
}

function checkDomainCart($params)
{
  try
  {
	$api_client = new ApiClient();
	if ($params['type'] != 'transfer')
	{
	  return;
	}

	// Check the rules for .NZ domains
	$should_renew = $api_client->ShouldBeRenewed($params['domain']);

	if ($should_renew)
	{
	  return;
	}

	return ['firstPaymentAmount' => 0.00];
  }
  catch (\Exception $e)
  {
	logActivity("[DOMAIN TRANSFER INTERCEPT] - ERROR ATTEMPTING TO UPDATE A CART PRICE - {$e->getMessage()}");
  }
}

function updateClientCart($params)
{
  try
  {
	$chargeNzDomain = [];
	if ($params['action'] != 'confdomains')
	{
	  return;
	}

	if (!array_key_exists('domains', $params) && count($params['domains']) == 0)
	{
	  return;
	}

	$api_client = new ApiClient();
	foreach ($params['domains'] as $num => $domainInfo)
	{
	  // If we have an eppvalue on our array, this is a transfer domain request
	  if ($domainInfo['eppenabled'] != true)
	  {
		$chargeNzDomain[$num] = 1;
		continue;
	  }
	  // We check if this domain is indeed a .nz domain.
	  $should_renew = $api_client->ShouldBeRenewed($domainInfo['domain']);

	  if ($should_renew)
	  {
		$chargeNzDomain[$num] = 1;
		continue;
	  }

	  $chargeNzDomain[$num] = 0;
	}

	return ['charge_nz_domain' => $chargeNzDomain];
  }
  catch (\Exception $e)
  {
	logActivity("[DOMAIN TRANSFER INTERCEPT] - ERROR ATTEMPTING TO UPDATE A CLIENT CART - {$e->getMessage()}");
  }

}

function shouldRemoveRegistrationPeriod($order_details): bool
{
  // Convert the order details from string to array
  $order_details = explode("<br>", str_replace("\n", '', $order_details));

  $is_transfer           = false;
  $should_not_be_renewed = false;
  $api_client            = new ApiClient();

  foreach ($order_details as $row)
  {
	[$title, $value] = explode(': ', $row, 2);

	// Check if this transaction is a domain transfer
	if ($title === 'Domain Registration' && $value === 'Transfer')
	{
	  $is_transfer = true;
	}
	// Check if this line has the domain and check if said domain should be renewed
	if ($title === 'Domain' && !$api_client->ShouldBeRenewed($value))
	{
	  $should_not_be_renewed = true;
	}
  }

  // We will remove the reg period when the process is a domain transfer and the domain do not have to be renewed
  return $is_transfer && $should_not_be_renewed;
}

function checkEmailToBeSent($params)
{
  if ($params['messagename'] !== 'Order Confirmation')
  {
	return;
  }

  try
  {
	$response           = [];
	$order_details_list = $params['mergefields']['order_details'];
	// Separate by item
	$order_details_list = explode("<br><br>", str_replace("\n", "", $order_details_list));

	foreach ($order_details_list as $key => $order_details)
	{
	  if (shouldRemoveRegistrationPeriod($order_details))
	  {
		$response[$key] = preg_replace('#Registration Period.*$#', '', $order_details);
	  }
	  else
	  {
		$response[$key] = $order_details;
	  }
	}
	// There is no need to add the "\n" again according to testing
	$response = implode('<br><br>', $response);

	return ['order_details' => $response];
  }
  catch (\Exception $e)
  {
	logActivity("[DOMAIN TRANSFER INTERCEPT] - ERROR ATTEMPTING TO EDIT THE CONFIRMATION EMAIL - {$e->getMessage()}");
  }
}

try
{
  $api_client      = new ApiClient();
  $api_credentials = $api_client->getApiCredentials();

  if ((array_key_exists('FreeNzDomainTransfer', $api_credentials)) && ($api_credentials['FreeNzDomainTransfer'] == 'on'))
  {
	// Hook for update invoice for transferred .NZ domains
	add_hook('InvoiceCreation', 2, 'check_nz_domain_rules');
	add_hook('OrderDomainPricingOverride', 1, 'checkDomainCart');
	add_hook('ClientAreaPageCart', 3, 'updateClientCart');
	add_hook('EmailPreSend', 4, 'checkEmailToBeSent');
  }
}
catch (\Exception $e)
{
  logActivity("[DOMAIN TRANSFER INTERCEPT] - Failed to add hooks");
}


/*
 *---------------------------------------------------------------------------
 * Hide domain lock option
 * --------------------------------------------------------------------------
 */

function getlocktlds()
{
  # This list is based on the CSC list, provided here:
  # https://docs.google.com/spreadsheets/d/1XjuE0VZdIS3gUJ8t4bpPjhP4SV6WAg0AyALPN_Uxe94
  # There are more domains that don't support it than do, however, hence the whitelist.
  return array(
	  'eu', 'com', 'info', 'mobi', 'net',
	  'org', 'pro', 'archi', 'bet', 'bio',
	  'black', 'blue', 'global', 'green',
	  'kim', 'lgbt', 'llc', 'lotto', 'organic',
	  'pet', 'pink', 'poker', 'promo', 'red',
	  'shiksha', 'ski', 'vote', 'voto', 'au',
	  'ac', 'at', 'be', 'ca', 'cc', 'ch', 'cn',
	  'cz', 'fi', 'fr', 'gr', 'hk', 'ie', 'io',
	  'jp', 'li', 'me', 'mx', 'nl', 'nu', 'pl',
	  'pt', 'se', 'sg', 'sh', 'si', 'tv', 'uk',
  );
}

add_hook('ClientAreaPrimarySidebar', 1, function(MenuItem $primarySidebar) {
  # Get the domain and extract the TLD
  $domain_obj = Menu::context('domain');
  $domain     = $domain_obj->domain;
  $tld        = substr($domain, strrpos($domain, ".") + 1);
  $locktlds   = getlocktlds();

  if (!in_array($tld, $locktlds))
  {
	# Remove as a sidebar option
	if (!is_null($primarySidebar->getChild('Domain Details Management')))
	{
	  $primarySidebar->getChild('Domain Details Management')
					 ->removeChild('Registrar Lock Status');
	}
  }
});

add_hook('ClientAreaPageDomainDetails', 1, function($vars) {
  $domain_obj = Menu::context('domain');
  $domain     = $domain_obj->domain;
  $tld        = substr($domain, strrpos($domain, ".") + 1);
  $locktlds   = getlocktlds();

  if (!in_array($tld, $locktlds))
  {
	# Remove as a client-area management option.
	$vars['managementoptions']['locking'] = false;
	$vars['lockstatus']                   = false;

	return $vars;
  }
});