<?php

namespace SiteHost\WHMCS;

use libphonenumber\NumberParseException;
use libphonenumber\PhoneNumberFormat;
use libphonenumber\PhoneNumberUtil;
use WHMCS\Database\Capsule;
/**
 * SiteHost API Client for WHMCS
 *
 * This class provides an interface for the main module file (sitehost.php) allowing it to talk to
 * the SiteHost domain API endpoints, while making minimal changes to the original main module file.
 *
 * @author SiteHost Developers <developers@sitehost.co.nz>
 */
class ApiClient
{
    const DEBUG_MODE = false;
    const SITEHOST_REGISTRAR_SLUG = 'sitehost';

    private $curl;
    private $server;
    private $key;
    private $client_id;

    private $response;
    private $error;

    // DNS record types that we disallow making changes to
    private $disallowedRecordTypes = ['SOA', 'NS', 'MXE', 'URL', 'FRAME'];

    // API version
    const API_VERSION = "1.2";

    /**
     * Constructor
     */
    public function __construct()
    {
        // Initialize a cURL session
        $this->curl = curl_init();
        curl_setopt($this->curl, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($this->curl, CURLOPT_SSL_VERIFYHOST, 2);
        curl_setopt($this->curl, CURLOPT_SSL_VERIFYPEER, 1);
        curl_setopt($this->curl, CURLOPT_TIMEOUT, 30);
        curl_setopt($this->curl, CURLOPT_CONNECTTIMEOUT, 10);
        curl_setopt($this->curl, CURLOPT_USERAGENT, 'SiteHost WHMCS API Client');
    }

    /**
     * Destructor
     */
    public function __destruct()
    {
        // Close the cURL session
        curl_close($this->curl);
    }

    /**
     * Processes all requests made by the module
     *
     * @param string $action
     * @param string $params
     *
     * @throws \Exception
     * @return mixed
     */
    public function call($action, $params)
    {
        $this->debug('[Function Call] ' . $action, $params);

        // Clear the previous response
        $this->response = null;

        // Check that a matching method for the action exists
        if (!method_exists($this, $action)) {
            throw new \Exception('The requested action has not been implemented.');
        }

        // Check that we have been passed the API server address
        if (!isset($params['api_server']) || $params['api_server'] === '') {
            throw new \Exception('The API server address has not been configured.');
        }

        // Store and unset the API server address
        $this->server = $params['api_server'];
        unset($params['api_server']);

        // Check that we have been passed an API key
        if (!isset($params['api_key']) || $params['api_key'] === '') {
            throw new \Exception('The API key has not been configured.');
        }

        // Store and unset the API key
        $this->key = $params['api_key'];
        unset($params['api_key']);

        // Look up the client ID
        $response = $this->sendRequest('/api/get_info', 'get');

        if ($response === false) {
            throw new \Exception('Failed to retrieve client ID: ' . $this->error);
        }

        $this->client_id = $response['client_id'];

        // Call the method
        $return = $this->$action($params);

        if (!is_null($this->response)) {
            $this->debug('[Function Response] ' . $action, $this->response);
        }

        return $return;
    }

    /**
     * Returns the value of the specified key from the last response received
     *
     * @param string $key
     *
     * @return string
     */
    public function getFromResponse($key)
    {
        // Return an empty string when the key cannot be found
        return (isset($this->response[$key]) ? $this->response[$key] : '');
    }

    /**
     * Returns the full response received for the last request
     *
     * @return mixed
     */
    public function getFullResponse()
    {
        return $this->response;
    }

    /**
     * Writes debugging data to a log file when DEBUG_MODE is set to true
     *
     * @param string $msg
     * @param string $data
     */
    private function debug($msg, $data)
    {
        if (self::DEBUG_MODE) {
            $log  = "\n--- " . $msg . ' ' . str_repeat('-', (95 - strlen($msg))) . "\n";

            ob_start();
            var_dump($data);
            $log .= ob_get_clean();

            file_put_contents('/tmp/sitehost-debug.log', $log, FILE_APPEND);
        }
    }

    /**
     * Sends a request to the API
     *
     * @param string $endpoint
     * @param string $method
     * @param array  $params
     *
     * @throws \Exception
     *
     * @return array|true
     */
    private function sendRequest($endpoint, $method, $params = array())
    {
        // Clear any previous error message
        $this->error = '';

        // Prepare the URL to call
        $url = $this->server . $this::API_VERSION . $endpoint . '.json?apikey=' . $this->key;

        // Set the request params
        $method = strtoupper($method);
        if ($method == 'GET') {
            $url .= '&' . http_build_query($params);
            curl_setopt($this->curl, CURLOPT_POST, 0);
        } else {
            curl_setopt($this->curl, CURLOPT_POST, 1);
            curl_setopt($this->curl, CURLOPT_POSTFIELDS, $params);
        }

        // Set the request URL
        curl_setopt($this->curl, CURLOPT_URL, $url);

        $this->debug('[API Request] ' . $method . ' ' . $endpoint, $params);

        // Execute the request
        $response = curl_exec($this->curl);

        // Check for any connection errors
        if (curl_errno($this->curl)) {
            throw new \Exception('Connection failed: ' . curl_error($this->curl));
        }

        $this->debug('[API Response] ' . $method . ' ' . $endpoint, $response);

        // Parse the JSON response
        $response = json_decode($response, true);

        // Check that we have a valid response
        if (!is_array($response) || !isset($response['msg'])) {
            throw new \Exception('Malformed response received.');
        }

        // Check that the request was successful
        if (!isset($response['status']) || !$response['status']) {
            // Strip the 'Error: ' prefix if present
            if (strpos($response['msg'], 'Error: ') === 0) {
                $response['msg'] = substr($response['msg'], strlen('Error: '));
            }

            $this->error = $response['msg'];
            return false;
        }

        // Return the response body if present otherwise just return true
        return (isset($response['return']) ? $response['return'] : true);
    }

    /**
     * Sends a request to the internal API for WHMCS
     *
     * @param string $command
     * @param array  $params
     *
     * @return mixed
     */
    private function sendInteralRequest($command, array $params = array())
    {
        try {
            // Check that the localAPI function is present
            if (!function_exists('localAPI')) {
                throw new \Exception('Local API function is not present.');
            }

            // Check that the required Capsule classes are present
            if (!class_exists('\Illuminate\Database\Capsule\Manager')) {
                throw new \Exception('Capsule DBAL is not present.');
            }

            // Look up an admin user to make the internal API request
            $admin = \Illuminate\Database\Capsule\Manager::table('tbladmins')->where('disabled', '=', 0)->first();

            // Check that we were able to find a valid admin user
            if (!is_object($admin) || !property_exists($admin, 'username')) {
                throw new \Exception('Unable to retrieve a valid admin user.');
            }

            // Send the API request to WHMCS
            $response = localAPI($command, $params, $admin->username);

            // Check that we received a valid response
            if (!is_array($response) || !isset($response['result'])) {
                throw new \Exception('Malformed response received from internal API.');
            }

            // Check that our request was successful
            if ($response['result'] !== 'success') {
                if (!isset($response['message'])) {
                    $response['message'] = 'An unknown error occurred when sending a request to the internal API.';
                }
                throw new \Exception($response['message']);
            }

            // Unset the result flag from the response
            unset($response['result']);

            // Return the rest of the response or true when there is no further response data
            return (!empty($response) ? $response : true);
        } catch (\Exception $e) {
            $this->error = $e->getMessage();
            return false;
        }
    }

    /**
     * Formats the given phone number as either three separate parts (for us) or one string (for WHMCS)
     *
     * Examples:
     *
     *    formatPhoneNumber('09 123 4567', 'NZ')    -->  ['Country' => '64', 'Area' => '9', 'Local' => '1234567']
     *
     *    formatPhoneNumber('64', '09', '1234567')  -->  '+64 (09) 1234567'
     *
     * @return string|array|false
     */
    private function formatPhoneNumber()
    {
        $args = func_get_args();
        $num_args = func_num_args();

        if ($num_args == 3) {
            // Format the number parts
            return vsprintf('+%s (%s) %s', preg_replace('/[^0-9]/', '', $args));
        } elseif ($num_args == 2) {
            // Set some default values for our result
            $result = array_fill_keys(['Country', 'Area', 'Local'], 0);

            try {
                $util = PhoneNumberUtil::getInstance();

                // Parse the given phone number using the two-letter country code
                $parsed = $util->parse($args[0], $args[1]);

                // Get the country code
                $result['Country'] = '+' . $parsed->getCountryCode();

                // Format the phone number as ITU-T E.123 international format
                $number = $util->format($parsed, PhoneNumberFormat::INTERNATIONAL);

                // Remove the country code
                if (strpos($number, $result['Country']) === 0) {
                    $number = trim(substr($number, strlen($result['Country'])));
                }

                // Split the number into parts
                $parts = preg_split('/[\s\-\)]/', $number, 2, PREG_SPLIT_NO_EMPTY);

                // Set the area and local parts of the number
                if (count($parts) == 2) {
                    $result['Area']  = $parts[0];
                    $result['Local'] = $parts[1];
                } else {
                    $result['Local'] = $parts[0];
                }

                // Strip any non-numeric chars
                $result = array_map(function ($value) {
                    return preg_replace('/[^0-9]/', '', $value);
                }, $result);
            } catch (NumberParseException $e) {
                $this->debug('Error parsing phone number: ' . $e->getMessage(), $args);
            }

            return $result;
        }

        return false;
    }

    /**
     * Returns info about the given contact or contact ID
     *
     * @param array|integer $contact
     *
     * @return array|false
     */
    private function getContact($contact)
    {
        // Check if we have been passed an array of contact details to search for a contact
        if (is_array($contact)) {
            $params = array(
                'client_id'    => $this->client_id,
                'query[name]'  => $contact['firstname'] . ' ' . $contact['lastname'],
                'query[email]' => $contact['email'],
            );

            $optional = array(
                'query[postal_address]' => 'address1',
                'query[suburb]'         => 'address2',
                'query[postcode]'       => 'postcode',
                'query[province]'       => 'state',
                'query[city]'           => 'city',
                'query[country]'        => 'country',
            );

            foreach ($optional as $k => $v) {
                if (isset($contact[$v]) && $contact[$v] !== '') {
                    $params[$k] = $contact[$v];
                }
            }

            if (isset($contact['phone'])) {
                $phone = $this->formatPhoneNumber($contact['phone'], $contact['country']);

                if ($phone !== false) {
                    $params['query[phone_local]'] = $phone['Local'];
                }
            }

            $response = $this->sendRequest('/srs/search_contacts', 'get', $params);

            if (is_array($response) && !empty($response)) {
                $contact = current($response);
                $contact = $contact['contact_id'];
            }
        }

        if (is_numeric($contact)) {
            // Retrieve the contact by ID
            $response = $this->sendRequest('/srs/get_contact', 'get', array(
                'client_id'  => $this->client_id,
                'contact_id' => $contact,
            ));

            if ($response !== false) {
                return $response;
            }
        }

        return false;
    }

    /**
     * Creates a contact using the given parameters
     *
     * @param array $contact
     *
     * @throws \Exception
     *
     * @return mixed
     */
    private function createContact(array $contact)
    {
        $params = array(
            'client_id'        => $this->client_id,
            'name'             => $contact['firstname'] . ' ' . $contact['lastname'],
            'email'            => $contact['email'],
            'postal_address'   => $contact['address1'],
            'postal_address2'  => "",
            'suburb'           => $contact['address2'],
            'city'             => $contact['city'],
            'country'          => $contact['country'],
            'params[PostCode]' => $contact['postcode'],
            'params[Province]' => $contact['state'],
        );

        $phone = $this->formatPhoneNumber($contact['phone'], $contact['country']);

        foreach (array('Phone', 'Fax', 'Mobile') as $num_type) {
            foreach ($phone as $k => $v) {
                $params['params[' . $num_type . '][' . $k . ']'] = $v;
            }
        }

        return $this->sendRequest('/srs/create_contact', 'post', $params);
    }

    /**
     * Returns the nameservers for the given domain
     *
     * @param string $domain
     *
     * @throws \Exception
     *
     * @return array
     */
    private function getDomainNameservers($domain)
    {
        $nameservers = $this->sendRequest('/srs/list_name_servers', 'get', array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($domain,IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
        ));

        if (!is_array($nameservers)) {
            throw new \Exception('Failed to retrieve nameservers: ' . $this->error);
        }

        return array_values($nameservers);
    }

    /**
     * Sets the given nameservers for a domain
     *
     * @param string $domain
     * @param array  $nameservers
     *
     * @return mixed
     */
    private function setDomainNameservers($domain, array $nameservers)
    {
        $nameservers = array_values($nameservers);

        $params = array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($domain,IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
        );

        for ($i = 0; $i < count($nameservers); $i++) {
            // Skip empty or null nameservers
            if (is_null($nameservers[$i]) || $nameservers[$i] === '') {
                continue;
            }

            $params['nameservers[' . $i . '][name]'] = $nameservers[$i];
        }

        return $this->sendRequest('/srs/add_name_servers', 'post', $params);
    }

    /**
     * Updates a glue record for a given domain
     *
     * @param string $domain
     * @param string $nameserver
     * @param mixed  $ip
     *
     * @throws \Exception
     *
     * @return mixed
     */
    private function updateGlueRecord($domain, $nameserver, $ip = false)
    {
        $params = array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($domain,IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
        );

        // Check that the nameserver ends with the domain
        if (substr($nameserver, -strlen($domain)) !== $domain) {
            throw new \Exception('Cannot register nameserver because it does not belong to this domain.');
        }

        // Get a list of current nameservers for the domain
        $current = $this->getDomainNameservers($domain);

        $i = 0;
        $exists = false;

        foreach ($current as $ns) {
            // Update the IP if this nameserver matches the one we're working with
            if ($ns['name'] == $nameserver) {
                if ($ip !== false) {
                    $ns['ipv4addr'] = $ip;
                    $ns['ipv6addr'] = '';
                    $exists = true;
                } else {
                    // We need to remove the nameserver if we haven't been given an IP
                    continue;
                }
            }

            // Append each nameserver
            $params['nameservers[' . $i . '][name]']     = $ns['name'];
            $params['nameservers[' . $i . '][ipv4addr]'] = $ns['ipv4addr'];
            $params['nameservers[' . $i . '][ipv6addr]'] = $ns['ipv6addr'];

            $i++;
        }

        // Add the nameserver if it didn't already exist above
        if (!$exists && $ip !== false) {
            $params['nameservers[' . $i . '][name]']     = $nameserver;
            $params['nameservers[' . $i . '][ipv4addr]'] = $ip;
            $params['nameservers[' . $i . '][ipv6addr]'] = '';
        }

        return $this->sendRequest('/srs/add_name_servers', 'post', $params);
    }

    /**
     * Creates a DNS zone for the given domain, if one does not already exist
     *
     * @param string $domain
     *
     * @throws \Exception
     *
     * @return mixed
     */
    private function createDnsZone($domain)
    {
        // Check if the zone already exists
        $zones = $this->sendRequest('/dns/search_domains', 'get', array(
            'client_id'     => $this->client_id,
            'query[domain]' => idn_to_ascii($domain,IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
        ));

        if (is_array($zones) && !empty($zones)) {
            return false;
        }

        // Create the new DNS zone
        $zone = $this->sendRequest('/dns/create_domain', 'post', array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($domain,IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
        ));

        if ($zone === false) {
            throw new \Exception('Failed to create DNS zone: ' . $this->error);
        }

        $sitehost_ns = array(
            'ns1.sitehost.co.nz',
            'ns2.sitehost.co.nz',
            'ns3.sitehost.co.nz',
        );

        // Add NS records to the zone
        foreach ($sitehost_ns as $ns) {
            $params = array(
                'client_id' => $this->client_id,
                'domain'    => idn_to_ascii($domain,IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
                'type'      => 'NS',
                'name'      => idn_to_ascii($domain,IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
                'content'   => $ns,
            );

            $record = $this->sendRequest('/dns/add_record', 'post', $params);

            if ($record === false) {
                throw new \Exception('Failed to add NS record: ' . $this->error);
            }
        }

        return $zone;
    }

    /**
     * Checks the availability of the given domain and optionally other TLDs and suggestions
     *
     * @param string  $domain
     * @param array   $additional_tlds
     * @param boolean $include_suggestions
     *
     * @throws \Exception
     *
     * @return array
     */
    private function checkDomainAvailability($domain, $additional_tlds = array(), $include_suggestions = false)
    {
        // Prepare the search params
        $search = array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($domain,IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
        );

        // Append the additional TLDs to check
        $tlds = array_values($additional_tlds);
        for ($i = 0; $i < count($tlds); $i++) {
            $search['additional_tlds[' . $i . ']'] = ltrim($tlds[$i], '.');
        }

        // Include suggestions for other similar domains
        if ($include_suggestions) {
            $search['suggest'] = '1';
        }

        // Send the request
        $response = $this->sendRequest('/srs/domain_tlds_available', 'get', $search);

        if (!is_array($response)) {
            throw new \Exception('Failed to check domain availability: ' . $this->error);
        }

        // Use the 'other' domains as our starting point for the result
        $domains = (isset($response['other']) ? $response['other'] : array());

        // Prepend the base domain and its state
        array_unshift($domains, array(
            'domain'    => $response['domain'],
            'available' => $response['available'],
        ));

        return $domains;
    }

    /**
     * Sends an email via WHMCS using a template defined in the WHMCS admin control panel
     *
     * @param string  $template
     * @param integer $related_id
     * @param array   $placeholders
     *
     * @return mixed
     */
    private function sendEmail($template, $related_id, array $placeholders = array())
    {
        return $this->sendInteralRequest('SendEmail', array(
            'messagename' => $template,
            'id'          => $related_id,
            'customvars'  => $placeholders,
        ));
    }


    //
    // All functions from this point onwards represent the actions called from the module using call()
    //

    private function Register($params)
    {
        // Prepare the registrant contact
        $reg_contact = $this->getContact($params['contacts']['registrant']);

        if ($reg_contact === false) {
            $reg_contact = $this->createContact($params['contacts']['registrant']);

            if ($reg_contact === false) {
                throw new \Exception('Failed to create a registrant contact: ' . $this->error);
            }
        }

        // Prepare the admin contact
        if (isset($params['contacts']['admin'])) {
            $admin_contact = $this->getContact($params['contacts']['admin']);

            if ($admin_contact === false) {
                $admin_contact = $this->createContact($params['contacts']['admin']);
            }
        }

        // Prepare the technical contact
        if (isset($params['contacts']['tech'])) {
            $tech_contact = $this->getContact($params['contacts']['tech']);

            if ($tech_contact === false) {
                $tech_contact = $this->createContact($params['contacts']['tech']);
            }
        }

        // Prepare the billing contact
        if (isset($params['contacts']['billing'])) {
            $billing_contact = $this->getContact($params['contacts']['billing']);

            if ($billing_contact === false) {
                $billing_contact = $this->createContact($params['contacts']['billing']);
            }
        }

        // Prepare the domain registration details
        $registration = array(
            'client_id'            => $this->client_id,
            'domain'               => $params['domain'],
            'term'                 => $params['years'] * 12,
            'registrant_contact'   => $reg_contact['ContactID'],
            'params[protect]'      => ($params['idprotection'] ? '1' : '0'),
        );

        // Set the admin contact ID
        if (isset($admin_contact) && is_array($admin_contact)) {
            $registration['params[AdminContact]'] = $admin_contact['ContactID'];
        }

        // Set the technical contact ID
        if (isset($tech_contact) && is_array($tech_contact)) {
            $registration['params[TechContact]'] = $tech_contact['ContactID'];
        }

        // Set the billing contact ID
        if (isset($billing_contact) && is_array($billing_contact))
        {
            $registration['params[BillingContact]'] = $billing_contact['ContactID'];
        }


        // Send the request to register the domain
        $registration = $this->sendRequest('/srs/create_domain', 'post', $registration);

        if ($registration === false) {
            throw new \Exception('Failed request domain registration: ' . $this->error);
        }

        // Set the nameservers for the domain
        $nameservers = $this->setDomainNameservers($params['domain'], $params['nameservers']);

        if ($nameservers === false) {
            throw new \Exception('Failed to set domain nameservers: ' . $this->error);
        }

        // Setup a DNS zone for the domain
        if ($params['dnsmanagement']) {
            $this->createDnsZone($params['domain']);
        }
    }

    private function Transfer($params)
    {
        // Prepare the registrant contact
        $reg_contact = $this->getContact($params['contacts']['registrant']);

        if ($reg_contact === false) {
            $reg_contact = $this->createContact($params['contacts']['registrant']);

            if ($reg_contact === false) {
                throw new \Exception('Failed to create a registrant contact: ' . $this->error);
            }
        }

        // Prepare the admin contact
        if (isset($params['contacts']['admin'])) {
            $admin_contact = $this->getContact($params['contacts']['admin']);

            if ($admin_contact === false) {
                $admin_contact = $this->createContact($params['contacts']['admin']);
            }
        }

        // Prepare the technical contact
        if (isset($params['contacts']['tech'])) {
            $tech_contact = $this->getContact($params['contacts']['tech']);

            if ($tech_contact === false) {
                $tech_contact = $this->createContact($params['contacts']['tech']);
            }
        }

        // Prepare the billing contact
        if (isset($params['contacts']['billing'])) {
            $billing_contact = $this->getContact($params['contacts']['billing']);

            if ($billing_contact === false) {
                $billing_contact = $this->createContact($params['contacts']['billing']);
            }
        }

        // Fall back to using the registrant contact if we have no admin contact
        if (!isset($admin_contact) || !is_array(($admin_contact))) {
            $admin_contact = $reg_contact;
        }

        // Fall back to using the admin contact if we have no technical contact
        if (!isset($tech_contact) || !is_array($tech_contact)) {
            $tech_contact = $admin_contact;
        }

        // Fall back to using the admin contact if we have no billing contact
        if (!isset($billing_contact) || !is_array($billing_contact)) {
            $billing_contact = $admin_contact;
        }

        // Check if this is a .NZ domain, if it is, we don't have to make a renewal
        $term = $this->ShouldBeRenewed($params['domain']) ? $params['years'] * 12 : null;

        // Prepare the transfer details
        $transfer = array(
            'client_id'             => $this->client_id,
            'domain'                => idn_to_ascii($params['domain'],IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
            'udai'                  => $params['eppcode'],
            'params[registrant_contact_id]' => $reg_contact['ContactID'],
            'params[admin_contact_id]'      => $admin_contact['ContactID'],
            'params[technical_contact_id]'  => $tech_contact['ContactID'],
            'params[billing_contact_id]'    => $billing_contact['ContactID'],
            'params[term]'                  => $term,
        );

        // Append the nameservers
        $nameservers = array_values($params['nameservers']);
        for ($i = 0; $i < count($nameservers); $i++) {
            // Skip empty or null nameservers
            if (is_null($nameservers[$i]) || $nameservers[$i] === '') {
                continue;
            }

            $transfer['params[nameservers][' . $i . '][name]'] = $nameservers[$i];
        }

        // Send the request to transfer the domain
        $transfer = $this->sendRequest('/srs/transfer_domain', 'post', $transfer);

        if ($transfer === false) {
            // Send an email to the client notifying them that we failed to initiate their domain transfer
            $this->sendEmail('Domain Transfer Failed', $params['domain_id'], array(
                'domain_transfer_failure_reason' => $this->error,
            ));

            throw new \Exception('Failed to initiate the domain transfer: ' . $this->error);
        }

        // Setup a DNS zone for the domain
        if ($params['dnsmanagement']) {
            $this->createDnsZone($params['domain']);
        }
    }

    private function Renew($params)
    {
        // Prepare the renewal details
        $renewal = array(
            'client_id'        => $this->client_id,
            'domain'           => idn_to_ascii($params['domain'],IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
            'term'             => $params['years'] * 12,
        );

        if ($params['idprotection']) {
            $renewal['options[privacy]'] = '1';
        }

        // Send the renewal request
        $renewal = $this->sendRequest('/srs/renew_domain', 'post', $renewal);

        if ($renewal === false) {
            throw new \Exception('Failed to request domain renewal: ' . $this->error);
        }
    }

    private function GetNameservers($params)
    {
        $nameservers = $this->getDomainNameservers($params['domain']);

        // Format the nameservers and append to our response
        for ($i = 0; $i < count($nameservers); $i++) {
            $this->response['nameserver' . ($i + 1)] = $nameservers[$i]['name'];
        }
    }

    private function SetNameservers($params)
    {
        // Extract the relevant 'nameserver' params we have been passed
        $nameservers = array_filter($params, function ($key) {
            return strpos($key, 'nameserver') === 0;
        }, ARRAY_FILTER_USE_KEY);

        $this->setDomainNameservers($params['domain'], $nameservers);
    }

    private function GetWhoisInformation($params)
    {
        // Retrieve the domain details to get the IDs for the current contacts
        $domain = $this->sendRequest('/srs/get_domain', 'get', array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($params['domain'],IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
        ));

        if ($domain === false) {
            throw new \Exception('Failed to retrieve domain details: ' . $this->error);
        }

        $contacts = array(
            'registrant' => $domain['RegistrantContactID'],
            'tech'       => $domain['TechnicalContactID'],
            'billing'    => $domain['BillingContactID'],
            'admin'      => $domain['AdminContactID'],
        );

        if ($contacts['billing'] === '0')
        {
           $contacts['billing'] = $contacts['admin'];
        }

        foreach ($contacts as $type => $id) {
            // Retrieve the details for each contact
            $contact = $this->getContact($id);

            // Format the phone numbers into a single string
            $phone = $this->formatPhoneNumber($contact['PhoneCountry'], $contact['PhoneArea'], $contact['PhoneLocal']);
            $fax   = $this->formatPhoneNumber($contact['FaxCountry'], $contact['FaxArea'], $contact['FaxLocal']);

            // Some crude splitting of first and last names :\
            $name = explode(' ', $contact['Name'], 2);
            $firstname = array_shift($name);
            $lastname  = (!empty($name) ? trim(current($name)) : '');

            $this->response[$type . '.firstname'] = $firstname;
            $this->response[$type . '.lastname']  = $lastname;
            $this->response[$type . '.company']   = ($contact['RegistrantName'] ?: $contact['Name']);
            $this->response[$type . '.email']     = $contact['Email'];
            $this->response[$type . '.address1']  = $contact['PostalAddress'];
            $this->response[$type . '.address2']  = $contact['Suburb'];
            $this->response[$type . '.city']      = $contact['City'];
            $this->response[$type . '.state']     = $contact['Province'];
            $this->response[$type . '.postcode']  = $contact['PostCode'];
            $this->response[$type . '.country']   = $contact['Country'];
            $this->response[$type . '.phone']     = $phone;
            $this->response[$type . '.fax']       = $fax;
        }
    }

    private function UpdateWhoisInformation($params)
    {
        // Retrieve the domain details to get the IDs for the current contacts
        $domain = $this->sendRequest('/srs/get_domain', 'get', array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($params['domain'],IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
        ));

        if ($domain === false) {
            throw new \Exception('Failed to retrieve current domain details: ' . $this->error);
        }

        $contact_ids = array(
            'registrant_contact_id' => $domain['RegistrantContactID'],
            'admin_contact_id'      => $domain['AdminContactID'],
            'technical_contact_id'  => $domain['TechnicalContactID'],
            'billing_contact_id'    => $domain['BillingContactID'],
        );

        // Map our contact types to the WHMCS types we will receive
        $contacts = array(
            'registrant_contact_id' => 'registrant',
            'admin_contact_id'      => 'admin',
            'technical_contact_id'  => 'tech',
            'billing_contact_id'    => 'billing',
        );

        foreach ($contacts as $type => $whmcs_type) {
            // Check that we have actually been passed a contact of this type
            if (!isset($params['contacts'][$whmcs_type])) {
                continue;
            }

            // Search for a contact with the details given
            $contact = $this->getContact($params['contacts'][$whmcs_type]);

            if (!is_array($contact)) {
                // We found no matching contact so we will need to create a new one
                $new_contact = $this->createContact($params['contacts'][$whmcs_type]);

                if ($new_contact !== false) {
                    $contact_ids[$type] = $new_contact['ContactID'];
                }
            } elseif ($contact['ContactID'] != $contact_ids[$type]) {
                // We found a matching contact but it doesn't match what is on the domain right now
                $contact_ids[$type] = $contact['ContactID'];
            } else {
                // Nothing to do here - the contact has not changed
                continue;
            }
        }

        // Update the contact IDs on the domain
        $response = $this->sendRequest('/srs/update_domain_contacts', 'post', array_merge(array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($params['domain'],IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
        ), $contact_ids));

        if ($response === false) {
            throw new \Exception('Failed to update WHOIS information: ' . $this->error);
        }
    }

    private function CheckAvailability($params)
    {
        // Default to checking availability of .nz domains when we haven't been given a valid list of TLDs
        $tlds = ['.nz'];

        if (is_array($params['tldsToSearch']) && !empty($params['tldsToSearch'])) {
            $tlds = $params['tldsToSearch'];
        }

        // Form a base domain to check using the search term and the first TLD in the list
        $domain = $params['searchTerm'] . array_shift($tlds);

        // Check the availability of the domain, using the remaining TLDs as additional TLDs to check
        $domains = $this->checkDomainAvailability($domain, $tlds);

        // Append each domain to the response
        $this->response['domains'] = array();
        foreach ($domains as $domain) {
            list($sld, $tld) = explode('.', $domain['domain'], 2);

            $this->response['domains'][] = array(
                'sld'           => $sld,
                'tld'           => $tld,
                'status'        => ($domain['available'] ? 'available' : 'registered'),
                'isPremiumName' => false,
            );
        }
    }

    private function GetSuggestions($params)
    {
        // Default to checking availability of .nz domains when we haven't been given a valid list of TLDs
        $tlds = ['.nz'];

        if (is_array($params['tldsToSearch']) && !empty($params['tldsToSearch'])) {
            $tlds = $params['tldsToSearch'];
        }

        // Form a base domain to check using the search term and the first TLD in the list
        $domain = $params['searchTerm'] . '.' . array_shift($tlds);

        // Check the availability of the domain, using the remaining TLDs as additional TLDs to check
        $domains = $this->checkDomainAvailability($domain, $tlds, true);

        // Append each domain to the response
        $this->response['domains'] = array();
        foreach ($domains as $domain) {
            list($sld, $tld) = explode('.', $domain['domain'], 2);

            // Exclude any domains that match the original SLD or are unavailable
            if ($sld == $params['searchTerm'] || !$domain['available']) {
                continue;
            }

            $this->response['domains'][] = array(
                'sld'           => $sld,
                'tld'           => $tld,
                'score'         => 0,
                'isPremiumName' => false,
            );
        }
    }

    private function GetLockStatus($params)
    {
        // Retrieve the details for the domain
        $domain = $this->sendRequest('/srs/get_domain', 'get', array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($params['domain'],IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
        ));

        if ($domain === false) {
            throw new \Exception('Failed to retrieve domain details: ' . $this->error);
        }

        // Check if the domain is locked
        if ($domain['Locked']) {
            $this->response['lockstatus'] = 'locked';
        } else {
            $this->response['lockstatus'] = 'unlocked';
        }
    }

    private function SetLockStatus($params)
    {
        // Determine whether or not we are locking or unlocking the domain
        if ($params['registrarlock']) {
            $action   = 'lock';
            $endpoint = '/srs/lock_domain';
        } else {
            $action   = 'unlock';
            $endpoint = '/srs/un_lock_domain';
        }

        // Send the request to lock or unlock the domain
        $response = $this->sendRequest($endpoint, 'post', array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($params['domain'],IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
        ));

        if ($response === false) {
            throw new \Exception('Failed to ' . $action . ' domain: ' . $this->error);
        }
    }

    private function GetDNSHostRecords($params)
    {
        // Create the DNS zone if it doesn't already exist
        $this->createDnsZone($params['domain']);

        // Retrieve the DNS records for the domain
        $records = $this->sendRequest('/dns/list_records', 'get', array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($params['domain'],IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
        ));

        if (!is_array($records)) {
            throw new \Exception('Failed to retrieve DNS records: ' . $this->error);
        }

        // Append each record to the response
        $this->response['records'] = array();
        foreach ($records as $record) {
            // Skip records of specific types
            if (in_array($record['type'], $this->disallowedRecordTypes)) {
                continue;
            }

            $this->response['records'][] = array(
                'name'    => $record['name'],
                'type'    => $record['type'],
                'address' => $record['content'],
                'mxpref'  => $record['prio'],
            );
        }
    }

    private function SetDNSHostRecords($params)
    {
        $domain = idn_to_ascii($params['domain'],IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46);

        // Retrieve the existing DNS records for the domain
        $current_records = $this->sendRequest('/dns/list_records', 'get', array(
            'client_id' => $this->client_id,
            'domain'    => $domain,
        ));

        if (!is_array($current_records)) {
            throw new \Exception('Failed to retrieve existing DNS records: ' . $this->error);
        }

        // Keep track of any errors we run into when creating/updating records
        $errors = [];

        // Iterate over the new set of records and find records to update or create
        foreach ($params['records'] as $new) {
            // Skip blank rows
            if ($new['hostname'] === '') {
                continue;
            }

            // Disallow changes to specific types of records
            if (in_array($new['type'], $this->disallowedRecordTypes)) {
                throw new \Exception('Records of type \'' . $new['type'] . '\' are not supported.');
            }

            // Transform '@' into the domain or append the domain when not already present
            if ($new['hostname'] == '@') {
                $new['hostname'] = $domain;
            } elseif (substr($new['hostname'], -strlen($domain)) !== $domain) {
                $new['hostname'] = $new['hostname'] . '.' . $domain;
            }

            // Set priority to 0 when we've been passed 'N/A'
            $new['priority'] = ($new['priority'] == 'N/A' ? 0 : $new['priority']);

            // Check this new record against each of the current records
            foreach ($current_records as $i => $current) {
                // Skip this record if the name or type do not match
                if ($new['hostname'] != $current['name'] || $new['type'] != $current['type']) {
                    continue 1;
                }

                // Check if the record content or priority have changed
                if ($new['address'] != $current['content'] || $new['priority'] != $current['prio']) {
                    $response = $this->sendRequest('/dns/update_record', 'post', array(
                        'client_id' => $this->client_id,
                        'domain'    => $domain,
                        'record_id' => $current['id'],
                        'type'      => $current['type'],
                        'name'      => $current['name'],
                        'content'   => $new['address'],
                        'prio'      => $new['priority'],
                    ));

                    if ($response === false) {
                        $errors[] = $current['name'] . ' (' . $current['type'] . '): ' . $this->error;
                    }
                }

                // Remove this record from the list of current records
                unset($current_records[$i]);

                // Skip to the next new record
                continue 2;
            }

            // Record does not already exist so we need to create a new record
            $response = $this->sendRequest('/dns/add_record', 'post', array(
                'client_id' => $this->client_id,
                'domain'    => $domain,
                'type'      => $new['type'],
                'name'      => $new['hostname'],
                'content'   => $new['address'],
                'prio'      => $new['priority'],
            ));

            if ($response === false) {
                $errors[] = $new['hostname'] . ' (' . $new['type'] . '): ' . $this->error;
            }
        }

        // Delete any remaining records in $current_records
        foreach ($current_records as $record) {
            // Disallow deletion of the SOA record and NS records
            if (in_array($record['type'], $this->disallowedRecordTypes)) {
                continue;
            }

            $this->sendRequest('/dns/delete_record', 'post', array(
                'client_id' => $this->client_id,
                'domain'    => $domain,
                'record_id' => $record['id'],
            ));
        }

        // Display any errors we might have encountered
        if (!empty($errors)) {
            $errors = '<ul><li>' . implode('</li><li>', $errors) . '</li></ul>';
            throw new \Exception('The following records could not be created or updated:<br>' . $errors);
        }
    }

    private function EnableIDProtection($params)
    {
        // Send a request to enable privacy protection
        $response = $this->sendRequest('/srs/enable_privacy_protection', 'post', array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($params['domain'],IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
            'reason'    => 'Requested via WHMCS',
        ));

        if ($response === false) {
            throw new \Exception('Failed to enable ID protection: ' . $this->error);
        }
    }

    private function DisableIDProtection($params)
    {
        // Send a request to disable privacy protection
        $response = $this->sendRequest('/srs/disable_privacy_protection', 'post', array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($params['domain'],IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
            'reason'    => 'Requested via WHMCS',
        ));

        if ($response === false) {
            throw new \Exception('Failed to disable ID protection: ' . $this->error);
        }
    }

    private function RequestEPPCode($params)
    {
        // Send a request for a new UDAI
        $response = $this->sendRequest('/srs/new_udai', 'get', array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($params['domain'],IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
        ));

        if ($response === false) {
            throw new \Exception('Failed to request EPP code: ' . $this->error);
        }

        // Set response to false to indicate that the code won't be returned and will be emailed instead
        $this->response['eppcode'] = false;
    }

    private function DeleteDomain($params)
    {
        // Send a request to cancel the domain
        $response = $this->sendRequest('/srs/cancel_domain', 'post', array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($params['domain'],IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
        ));

        if ($response === false) {
            throw new \Exception('Failed to request domain cancellation: ' . $this->error);
        }

        // Remove the DNS zone
        $this->sendRequest('/dns/delete_domain', 'post', array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($params['domain'],IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
        ));
    }

    private function RegisterNameserver($params)
    {
        if (!isset($params['ip']) || $params['ip'] === '') {
            throw new \Exception('Please specify a valid IP address for the nameserver.');
        }

        $response = $this->updateGlueRecord($params['domain'], $params['nameserver'], $params['ip']);

        if ($response === false) {
            throw new \Exception('Failed to register nameserver: ' . $this->error);
        }
    }

    private function ModifyNameserver($params)
    {
        if (!isset($params['newip']) || $params['newip'] === '') {
            throw new \Exception('Please specify a new IP address to modify the nameserver.');
        }

        $response = $this->updateGlueRecord($params['domain'], $params['nameserver'], $params['newip']);

        if ($response === false) {
            throw new \Exception('Failed to modify registered nameserver: ' . $this->error);
        }
    }

    private function DeleteNameserver($params)
    {
        $response = $this->updateGlueRecord($params['domain'], $params['nameserver']);

        if ($response === false) {
            throw new \Exception('Failed to delete registered nameserver: ' . $this->error);
        }
    }

    private function GetDomainInfo($params)
    {
        // Set some default response values
        $response = array(
            'expirydate'      => '0000-00-00',
            'active'          => false,
            'expired'         => false,
            'transferredaway' => false,
        );

        // Retrieve the details for the domain
        $domain = $this->sendRequest('/srs/get_domain', 'get', array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($params['domain'],IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
        ));

        if ($domain !== false) {
            if ($domain['DateRegistered'] == '0000-00-00 00:00:00' || $domain['DateBilledUntil'] == '0000-00-00 00:00:00')
            {
                // Domain has not been fully updated yet.
                $this->response = false;
                return;
            }
            else
            {
                $expiry                 = strtotime($domain['DateBilledUntil']);
                $response['expirydate'] = date('Y-m-d', $expiry);
                $response['active']     = true;
                $response['expired']    = ($expiry < time());
            }
        }
        else
        {
            if (isset($this->error) && strpos($this->error, "The specified domain doesnt exist") === false)
            {
                $response['error'] = $this->error;
            }
            else
            {
                // The domain must have been transferred away when we can't find it anymore
                $response['transferredaway'] = true;
            }
        }

        $this->response = $response;
    }

    private function CheckDomainTransfer($params)
    {
        // Set some default response values
        $response = array(
            'transfercomplete' => false,
            'transferfailed'   => false,
            'expirydate'       => '0000-00-00',
            'failurereason'    => '',
        );

        // Retrieve the details for the domain
        $domain = $this->sendRequest('/srs/get_domain', 'get', array(
            'client_id' => $this->client_id,
            'domain'    => idn_to_ascii($params['domain'],IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46),
        ));

        if ($domain !== false) {
            // Set the expiry date
            $response['expirydate'] = date('Y-m-d', strtotime($domain['DateBilledUntil']));

            // Check if we have a transfer state set
            $transfer_state = (isset($domain['TransferStatus']) ? strtolower($domain['TransferStatus']) : '');

            // Check the transfer state, otherwise fallback to using the actual domain state
            if (strpos($transfer_state, 'complete') !== false || $domain['State'] == 'Active') {
                $response['transfercomplete'] = true;
            } elseif (strpos($transfer_state, 'fail') !== false || $domain['State'] == 'Available') {
                $response['transferfailed'] = true;
            }
        } else {
            // When we can't find the domain, we can try checking its WHOIS data
            $whois = $this->sendRequest('/srs/whois', 'get', array('domain' => idn_to_ascii($params['domain'],IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46)));

            if ($whois !== false) {
                // Set the expiry date if we have one
                if (isset($whois['BilledUntil'])) {
                    $response['expirydate'] = date('Y-m-d', strtotime($whois['BilledUntil']));
                }

                // Check the state to figure out the transfer state
                if (isset($whois['Status'])) {
                    if ($whois['Status'] == 'Active') {
                        $response['transfercomplete'] = true;
                    } elseif ($whois['Status'] == 'Available') {
                        $response['transferfailed'] = true;
                    }
                }
            }
        }

        $this->response = $response;
    }

	/**
	 * Unsupported functions that we cannot implement:
	 *
	 * private function ReleaseDomain($params) {}
	 */

	/**
	 * Check if a domain should be renewed, a domain should be renewed if:
	 * - it is not a .NZ domain, or
	 * - it is a .NZ domain and about to expire within 21 days
	 *
	 * @param string $domain
	 *
	 * @return boolean
	 * @throws \Exception
	 */
	function ShouldBeRenewed(string $domain): bool
	{
		if (empty($domain))
		{
			return true;
		}
		$domain = strtolower($domain);

		// Check if this domain is a .nz domain
		$is_nz_domain = substr($domain, -3) === '.nz';
		if (!$is_nz_domain)
		{
			return true;
		}

		$api_credentials = $this->getApiCredentials();

		if (empty($api_credentials))
		{
			throw new \Exception('Cant get the api credentials.');
		}

        // Check if this registrar can change the price for the renewal
		if (!array_key_exists('FreeNzDomainTransfer', $api_credentials))
		{
			return true;
		}

		// If it can, check if is on or off
		if ($api_credentials['FreeNzDomainTransfer'] != 'on')
		{
			return true;
		}

		$this->key    = $api_credentials['APIKey'];
		$this->server = $api_credentials['APIServer'];

		$whois = $this->sendRequest('/srs/whois', 'GET', ['domain' => idn_to_ascii($domain,IDNA_DEFAULT,INTL_IDNA_VARIANT_UTS46)]);

		// Check if the domain has at least more than 21 days before his expiry date
		$domain_valid_at = strtotime($whois['BilledUntil']);
		$minimum_term = strtotime('+21 days');

		// Check if still has more than 21 days left for the renewal
		if ($domain_valid_at <= $minimum_term)
		{
			return true;
		}

		return false;
	}

	public function getApiCredentials(): array
	{
		// Get the API params
		$sitehost_info = Capsule::table('tblregistrars')
		                        ->where('registrar', self::SITEHOST_REGISTRAR_SLUG)
		                        ->get();
		$api_params    = [];
		foreach ($sitehost_info as $info)
		{
			$api_params[$info->setting] = Decrypt($info->value);
		}

		return $api_params;
	}
}
